package cs108;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Queue;

public final class BurrowsWheelerTransform {
    private BurrowsWheelerTransform() {}

    // Variante 1 du calcul des rotations : au moyen d'une queue
    private static Queue<Character> stringToQueue(String s) {
        Queue<Character> q = new ArrayDeque<>(s.length());
        for (int i = 0; i < s.length(); ++i)
            q.add(s.charAt(i));
        return q;
    }

    private static String queueToString(Queue<Character> q) {
        StringBuilder b = new StringBuilder(q.size());
        q.forEach(b::append);
        return b.toString();
    }

    private static List<String> allRotations(String s) {
        List<String> rs = new ArrayList<>(s.length());
        Queue<Character> q = stringToQueue(s);
        while (rs.size() < s.length()) {
            rs.add(queueToString(q));
            q.add(q.remove());
        }
        return rs;
    }

    // Variante 2 du calcul des rotations : au moyen d'un bâtisseur de chaîne
    private static List<String> allRotations2(String s) {
        List<String> rs = new ArrayList<>(s.length());
        while (rs.size() < s.length()) {
            rs.add(s);
            s = new StringBuilder(s.length())
                    .append(s, 1, s.length())
                    .append(s, 0, 1)
                    .toString();
        }
        return rs;
    }

    public static Pair<Integer, String> forward(String s) {
        if (s.isEmpty())
            throw new IllegalArgumentException("empty input string");

        List<String> rs = allRotations2(s);
        Collections.sort(rs);
        StringBuilder b = new StringBuilder(s.length());
        rs.forEach(r -> b.append(r.charAt(r.length() - 1)));
        return new Pair<>(rs.indexOf(s), b.toString());
    }

    public static String backward(Pair<Integer, String> p) {
        if (! (0 <= p.first() && p.first() < p.second().length()))
            throw new IndexOutOfBoundsException("invalid index " + p.first()
            + " (should be between 0 and " + (p.second().length() - 1) + ")");

        String s = p.second();
        List<String> rs = new ArrayList<>(Collections.nCopies(s.length(), ""));
        for (int i = 0; i < s.length(); ++i) {
            for (int j = 0; j < s.length(); ++j)
                rs.set(j, s.charAt(j) + rs.get(j));
            Collections.sort(rs);
        }
        return rs.get(p.first());
    }
}
